;; Generated package description from company.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "company" "1.1.0.0.20260721.0" "Modular text completion framework" '((emacs "26.1") (posframe "1.5.1")) :commit "1cc907ac9e46ae4209eb5a341131787e0c678406" :maintainer '("Dmitry Gutov" . "dmitry@gutov.dev") :keywords '("abbrev" "convenience" "matching") :url "http://company-mode.github.io/")
