((nil . ((indent-tabs-mode . nil)
         (fill-column . 80)
         (sentence-end-double-space . t)
         (emacs-lisp-docstring-fill-column . 75)
         (project-vc-merge-submodules . nil)))
 (makefile-mode . ((indent-tabs-mode . t)))
 (texinfo-mode . ((fill-column . 70))))
